/**********************************************************************
 
	Copyright (C) 2003 Hirohisa MORI <joshua@nichibun.ac.jp>
 
	This program is free software; you can redistribute it 
	and/or modify it under the terms of the GLOBALBASE 
	Library General Public License (G-LGPL) as published by 

	http://www.globalbase.org/
 
	This program is distributed in the hope that it will be 
	useful, but WITHOUT ANY WARRANTY; without even the 
	implied warranty of MERCHANTABILITY or FITNESS FOR A 
	PARTICULAR PURPOSE.

**********************************************************************/


#include	<sys/wait.h>
#include	<sys/time.h>
#include	<sys/types.h>
#include	<unistd.h>
#include	<stdlib.h>
#include	<stdio.h>
#include	<signal.h>
#include	"init.h"
#include	"memory_debug.h"
#include	"machine/err.h"
#include	"task.h"
#include	"utils.h"
#include	"machine/fork_lock.h"
#include	"lock_level.h"

typedef void * _TKEY;
extern int task_flags;
void ufr_close_all();
int ignore_child_flag;
void init_posix_dependent();

#define DTI_LENGTH		10000


typedef struct task_block {
	struct task_block *	next;
	int			cmd;
#define TBC_EXEC		1
#define TBC_FINISH		2
	HTID			hard_tid;
	int			soft_tid;
	int			pri;
	int			data;
	void 			(*func)();
	unsigned int		finish_time;

	char *			file;
	int			line;
} TASK_BLOCK;

typedef struct dti {
	TASK_BLOCK *		tb;
	void			(*func)();
	int			msg;
	int			msg2;
} DTI;

void aboat_resolve_request_init();
void init_wd();
void setup_main_thread();
int _create_task(void (*func)(),int data,int pri);
int task_init;
void _task_timeout();
int task_nos_limit;

DTI * debug_task_info;
extern HTID (*get_tid_ptr)();

#define TASK_IDLE_LIMIT		3600
#define TASK_BLOCK_HASH		23
#define TASK_NOS_LIMIT		400

int task_cnt;


extern int pri_table[PRI_MAX];

#define MAX_TEST 10
int test_tid[MAX_TEST];
int test_tno[MAX_TEST];

TASK_BLOCK *	task_hash[TASK_BLOCK_HASH];
D_SEM		task_lock;
int		soft_tid;

int
debug_th_timeout_check(int time)
{
  return 0;
}

void
test_no(int no)
{
int tid;
int i;
	tid = get_tid();
	for ( i = 0 ; i < MAX_TEST ; i ++ )
		if ( test_tid[i] == tid ) {
			test_tno[i] = no;
			return;
		}
	for ( i = 0 ; i < MAX_TEST ; i ++ )
		if ( test_tid[i] == 0 ) {
			test_tid[i] = tid;
			test_tno[i] = no;
			return;
		}
}

void
print_tno(char * str)
{
int i;
	printf("%s ",str);
	for ( i = 0 ; i < MAX_TEST ; i ++ ) {
		if ( test_tid[i] == 0 )
			continue;
		printf("(%i-%i)",test_tid[i],test_tno[i]);
	}
	printf("\n");
}


void
sleep_sec(int x)
{
	sleep(x);
}

#ifdef DEADLOCK_CHECK
D_SEM sem_list;

void
lock_check_task()
{
D_SEM s;
	sleep_sec(20);
	for ( ; ; ) {
		printf("sem");
		for ( s = sem_list ; s ; s = s->next ) {
			if ( s->tid == 0 )
				continue;
			printf("(%x-%i)",s,s->tid);
		}
		printf("\n");
		sleep_sec(10);
	}
}
#endif



void
init_task(int flags)
{
int i;

	if ( task_init )
		return;
	task_init = 1;

	get_tid_ptr = _get_tid;
	aboat_resolve_request_init();

	debug_task_info = malloc(sizeof(DTI)*DTI_LENGTH);
	for ( i = 0 ; i < DTI_LENGTH ; i ++ ) {
		debug_task_info[i].tb = 0;
		debug_task_info[i].func = 0;
		debug_task_info[i].msg = 0;
		debug_task_info[i].msg2 = 0;
	}

	set_alarm_mask(flags);
	init_thread_block();
	sw_init();

	task_lock = xx_new_lock(LL_TASK,__FILE__,__LINE__);
	_lock_task(task_lock,__FILE__,__LINE__);
	_create_task(_task_timeout,0,1);
	_unlock_task(task_lock,"init_task",__FILE__,__LINE__);

	setup_main_thread();
	task_init = 2;

	init_wd();

	init_fork_lock();

	init_u_fileroot();

	init_posix_dependent();
}


void
set_t_msg(int msg)
{
int tid;
	tid = get_tid();
	debug_task_info[tid % DTI_LENGTH].msg = msg;
}


void
set_t_msg2(int msg)
{
int tid;
	tid = get_tid();
	debug_task_info[tid % DTI_LENGTH].msg2 = msg;
}

int
get_pri(int tid)
{
struct sched_param sp;
int i;
int ret;
int p;
int key;
HTID _tid;
TASK_BLOCK *b;
	if ( tid == 0 )
		_tid = _get_tid();
	else	_tid = get_hard_tid(tid);
	errno = 0;
	if ( SCHED_POLICY == SCHED_OTHER ) {
		_lock_task(task_lock,__FILE__,__LINE__);
		key = _tid.tid % TASK_BLOCK_HASH;
		if ( key < 0 )
			key = -key;
		for ( b = task_hash[key] ; b ; b = b->next )
			if ( cmp_htid(b->hard_tid,_tid) == 0 ) {
				ret = b->pri;
				_unlock_task(task_lock,"get_pri",__FILE__,__LINE__);
				return ret;
			}
		er_panic("get_pri(OTHER)");
	}
	ret = pthread_getschedparam((pthread_t)_tid.tid,&p,&sp);
	switch ( ret ) {
	case 0:
		break;
	default:
		printf("cange_priority(1)%i\n",ret);
	}
	if ( errno )
	  perror("get_pri");
	for ( i = 0; i < PRI_MAX ; i ++ )
		if ( pri_table[i] == sp.sched_priority )
			return i;
	er_panic("invalid pri\n");
	return 0;
}

void
_change_pri(HTID hard_tid,int pri)
{
struct sched_param sp;
int ret;
int p;
TASK_BLOCK *b;
int key;
	if ( hard_tid.valid == 0 )
		hard_tid = _get_tid();
	key = hard_tid.tid % TASK_BLOCK_HASH;
	if ( key < 0 )
		key = -key;
	if ( pri < 0 )
		pri = 0;
	if ( pri >= PRI_MAX )
		pri = PRI_MAX-1;
	for ( b = task_hash[key] ; b ; b = b->next )
		if ( cmp_htid(b->hard_tid,hard_tid) == 0 ) {
			b->pri = pri;
			break;
		}
	if ( SCHED_POLICY != SCHED_OTHER ) {
		ret = pthread_getschedparam((pthread_t)hard_tid.tid,&p,&sp);
		switch ( ret ) {
		case 0:
			break;
		default:
			printf("cange_priority(1)%i\n",ret);
		}
		sp.sched_priority = pri_table[pri];
		ret = pthread_setschedparam((pthread_t)hard_tid.tid,
					SCHED_POLICY,&sp);
		switch ( ret ) {
		case 0:
			return;
		default:
			printf("cange_priority(2)%i\n",ret);
		}
	}
}

void
change_pri(int tid,int pri)
{
HTID hard_tid;
	if ( tid ) {
		hard_tid = get_hard_tid(tid);
		if ( hard_tid.valid == 0 )
			return;
	}
	else {
		hard_tid.valid = 0;
		hard_tid.tid = 0;
	}
	_lock_task(task_lock,__FILE__,__LINE__);
	_change_pri(hard_tid,pri);
	_unlock_task(task_lock,"change_pri",__FILE__,__LINE__);
}

int
_create_task(void (*func)(),int data,int pri)
{
pthread_t tid;
int ret;
void *(*f)();
int * d;
pthread_attr_t tattr;
struct sched_param param;
int policy;


	f = (void *(*)())func;
	d = malloc(sizeof(int));
	*d = data;
/*
	if ( task_cnt >= TASK_NOS_LIMIT ) {
		printf("TASK NOS LIMIT\n");
		return ESYS_AGAIN;
	}
*/

	ret = pthread_attr_init(&tattr);

	pthread_attr_setstacksize(&tattr,0x10000);

	if ( pri < 0 ) {
/*
		policy = SCHED_SUPER_POLICY;
		pri = 0;
*/
		policy = SCHED_POLICY;
		pri = PRI_MAX-1;
	}
	else	policy = SCHED_POLICY;
	ret = pthread_attr_setschedpolicy(&tattr,policy);
	if ( policy != SCHED_OTHER ) {
		pthread_attr_getschedparam(&tattr,&param);
		if ( pri >= PRI_MAX )
			pri = PRI_MAX-1;
		if ( pri < 0 )
			pri = 0;
		param.sched_priority = pri_table[pri];
		ret = pthread_attr_setschedparam(&tattr,&param);
		if ( ret )
			er_panic("create_task(set_schedparam)");
	}
	ret = pthread_attr_setdetachstate(&tattr,PTHREAD_CREATE_DETACHED);
	if ( ret )
		er_panic("create_task(detatch)");
	ret = pthread_create(&tid,&tattr,f,d);
/*
	fprintf(stderr,"tid = %i\n",tid);
*/
	switch ( ret ) {
	case 0:
/*
		fprintf(stderr,"OK\n");
*/
		break;
	case ESYS_AGAIN:
		fprintf(stdout,"EAGAIN %i\n",(int)getpid());
		break;
	case ESYS_INVAL:
		fprintf(stdout,"EINVAL %i\n",(int)getpid());
		break;
	default:
		fprintf(stdout,"default %i\n",(int)getpid());
		break;
	}
	return ret;
}

int
_GET_TKEY(_TKEY d)
{
int ret;
int * ptr;
	ptr = d;
	ret = *ptr;
	free(d);
	return ret;
}

/*
void
rw_lock_init(RW_LOCK * rwl,int level)
{
	rwl->lock = new_lock(level);
	rwl->cnt = 0;
}

void
read_lock(RW_LOCK * rwl)
{
struct timeval tm;
	lock_task(rwl->lock);
	for ( ; rwl->cnt < 0 ; ) {
		unlock_task(rwl->lock,"read_lock(1)");
		tm.tv_sec = 0;
		tm.tv_usec = 10000;
		select(0,0,0,0,&tm);
		lock_task(rwl->lock);
	}
	rwl->cnt ++;
	unlock_task(rwl->lock,"read_lock(2)");
}

void
write_lock(RW_LOCK * rwl)
{
struct timeval tm;
	lock_task(rwl->lock);
	for ( ; rwl->cnt != 0 ; ) {
		unlock_task(rwl->lock,"write_lock(1)");
		tm.tv_sec = 0;
		tm.tv_usec = 10000;
		select(0,0,0,0,&tm);
		lock_task(rwl->lock);
	}
	rwl->cnt = -1;
	unlock_task(rwl->lock,"write_lock(2)");
}

void
rwl_unlock(RW_LOCK * rwl)
{
	lock_task(rwl->lock);
	if ( rwl->cnt > 0 )
		rwl->cnt --;
	else if ( rwl->cnt < 0 )
		rwl->cnt = 0;
	else	er_panic("rwl_unlock(1)");
	unlock_task(rwl->lock,"rwl_unlock(1))");
}


void
init_fifo(FIFO * f)
{
	f->head = f->tail = 0;
	f->lock = new_lock();
	f->length = 0;
}


void
insert_fifo(FIFO * f,FIFO_EL * e)
{
	lock_task(f->lock);
	e->next = 0;
	if ( f->head == 0 )
		f->head = f->tail = e;
	else {
		f->tail->next = e;
		f->tail = e;
	}
	f->length ++;
	unlock_task(f->lock,"insert_fifo");
}


FIFO_EL *
delete_fifo(FIFO * f)
{
FIFO_EL * ret;
	lock_task(f->lock);
	if ( f->head == 0 ) {
		ret = 0;
	}
	else {
		ret = f->head;
		f->head = ret->next;
		if ( f->head == 0 )
			f->tail = 0;
		f->length --;
	}
	unlock_task(f->lock,"delete_fifo");
	return ret;
}
*/

HTID
_get_tid()
{
HTID ret;
	ret.tid = (int)pthread_self();
	ret.valid = 1;
	return ret;
}


void
_insert_task_hash(TASK_BLOCK * b)
{
int key;

	key = b->hard_tid.tid % TASK_BLOCK_HASH;
	if ( key < 0 )
		key = -key;
	b->next = task_hash[key];
	task_hash[key] = b;

	if ( debug_task_info ) {
	DTI * d;
		d = &debug_task_info[b->soft_tid % DTI_LENGTH];
		d->tb = b;
		d->func = b->func;
	}
}

void
_delete_task_hash(TASK_BLOCK * b)
{
int key;
TASK_BLOCK ** bp;

	key = b->hard_tid.tid % TASK_BLOCK_HASH;
	if ( key < 0 )
		key = -key;
	bp = &task_hash[key];
	for ( ; *bp ; bp = &(*bp)->next )
		if ( *bp == b ) {
			*bp = b->next;
			break;
		}
}

int
_get_new_tid()
{
int tid;
TASK_BLOCK * b;
int key;

next:
	tid = soft_tid + 1;
	if ( tid <= 1 )
		tid = 2;
	soft_tid = tid;
	for ( key = 0 ; key < TASK_BLOCK_HASH ; key ++ )
		for ( b = task_hash[key] ; b ; b = b->next )
			if ( b->soft_tid == tid )
				goto next;
	return tid;
}

TASK_BLOCK *
_get_free_block()
{
int key;
TASK_BLOCK * b;
	for ( key = 0 ; key < TASK_BLOCK_HASH ; key ++ )
		for ( b = task_hash[key] ; b ; b = b->next )
			if ( b->soft_tid == 0 && b->cmd != TBC_FINISH )
				return b;
	return 0;
}


void
_task_grue(_TKEY d)
{
TASK_BLOCK b;
TASK_BLOCK * ctb;
int er;


	pthread_setcancelstate(PTHREAD_CANCEL_DISABLE,0);

	set_alarm_mask(INI_DONTWAITCHI);

	if ( (task_flags&INI_THREAD) == INI_DONTWAITCHI )
		set_ignore_chld();
	ctb = (TASK_BLOCK*)_GET_TKEY(d);
	b = *ctb;
	free(ctb);
	b.hard_tid = _get_tid();
	b.finish_time = 0;
	_lock_task(task_lock,__FILE__,__LINE__);
	_insert_task_hash(&b);
	task_cnt ++;
	_unlock_task(task_lock,"_task_grue",__FILE__,__LINE__);
	for ( ; ; ) {

		if ( debug_task_info ) {
		DTI * d;
			d = &debug_task_info[b.soft_tid % DTI_LENGTH];
			d->tb = &b;
			d->func = b.func;
		}

		(*b.func)(b.data);

		close_thread_area();

		_lock_task(task_lock,__FILE__,__LINE__);
		b.soft_tid = 0;
		b.finish_time = get_xltime();
	re_sleep:
		t_sleep_task((int)&b,task_lock,0);

		_lock_task(task_lock,__FILE__,__LINE__);
		switch ( b.cmd ) {
		case TBC_EXEC:
			if ( b.soft_tid == 0 )
				goto re_sleep;
			break;
		case TBC_FINISH:
			_delete_task_hash(&b);
			task_cnt --;
			_unlock_task(task_lock,"_task_grue",__FILE__,__LINE__);
			er = pthread_detach((pthread_t)b.hard_tid.tid);
			pthread_exit(0);
			return;
		default:
			er_panic("_task_grue(1)");
		}
		_unlock_task(task_lock,"_task_grue",__FILE__,__LINE__);
	}
}

int 
xx_create_task(void (*func)(),int data,int pri,char * __f,int __l)
{
TASK_BLOCK * b;
int ret;
int _ret;


	_lock_task(task_lock,__FILE__,__LINE__);
retry:
	b = _get_free_block();
	if ( b ) {
		ret = b->soft_tid = _get_new_tid();
		b->data = data;
		b->func = func;
		b->finish_time = 0;
		b->cmd = TBC_EXEC;
		b->pri = pri;
		b->file = __f;
		b->line = __l;
		_change_pri(b->hard_tid,pri);
		t_wakeup_task((int)b,0);
	}
	else {
		b = malloc(sizeof(*b));
		ret = b->soft_tid = _get_new_tid();
		b->data = data;
		b->func = func;
		b->cmd = TBC_EXEC;
		b->pri = pri;
		b->file = __f;
		b->line = __l;
		_ret = _create_task(_task_grue,(int)b,pri);
		if ( _ret == ESYS_AGAIN ) {
			free(b);
			sleep_sec(1);
			goto retry;
		}
	}
	_unlock_task(task_lock,"create_task",__FILE__,__LINE__);
	return ret;
}

int
get_tid()
{
HTID hard_tid;
int soft_tid;
int key;
TASK_BLOCK * b;
	hard_tid = _get_tid();
	_lock_task(task_lock,__FILE__,__LINE__);
	key = hard_tid.tid % TASK_BLOCK_HASH;
	if ( key < 0 )
		key = -key;
	for ( b = task_hash[key] ; b ; b = b->next )
		if ( cmp_htid(b->hard_tid,hard_tid) == 0 ) {
			soft_tid = b->soft_tid;
			_unlock_task(task_lock,"get_tid",__FILE__,__LINE__);
			return soft_tid;
		}
	er_panic("get_tid");
	return 0;
}

void
_task_timeout()
{
unsigned int n;
unsigned int key;
TASK_BLOCK * b;
int cnt;
	for ( ; ; ) {
		sleep_sec(20);
		n = get_xltime();
		cnt = 0;
		_lock_task(task_lock,__FILE__,__LINE__);
		for ( key = 0 ; key < TASK_BLOCK_HASH ; key ++ )
			for ( b = task_hash[key] ; b ; b = b->next ) {
				cnt ++;
				if ( b->soft_tid )
					continue;
				if ( task_cnt >= TASK_NOS_LIMIT )
					goto finish;
				if ( n - b->finish_time <
						TASK_IDLE_LIMIT )
					continue;
			finish:
				b->cmd = TBC_FINISH;
				t_wakeup_task((int)b,0);
			}
		_unlock_task(task_lock,"_task_timeout",__FILE__,__LINE__);
	}
}

HTID
get_hard_tid(int tid)
{
int key;
TASK_BLOCK * b;
HTID ret;
	_lock_task(task_lock,__FILE__,__LINE__);
	for ( key = 0 ; key < TASK_BLOCK_HASH ; key ++ )
		for ( b = task_hash[key] ; b ; b = b->next )
			if ( b->soft_tid == tid ) {
				ret = b->hard_tid;
				_unlock_task(task_lock,"get_hard_tid",__FILE__,__LINE__);
				return ret;
			}
	_unlock_task(task_lock,"get_hard_tid",__FILE__,__LINE__);
	ret.valid  = 0;
	ret.tid = 0;
	return ret;
}

int
thread_exist(int tid)
{
	if ( get_hard_tid(tid).valid == 0 )
		return -1;
	return 0;
}

void
setup_main_thread()
{
TASK_BLOCK * b;
	_lock_task(task_lock,__FILE__,__LINE__);
	b = malloc(sizeof(*b));
	b->soft_tid = _get_new_tid();
	b->hard_tid.tid = (unsigned int)pthread_self();
	b->hard_tid.valid = 1;
	b->data = 0;
	b->func = 0;
	b->cmd = TBC_EXEC;
	_insert_task_hash(b);
	_unlock_task(task_lock,"setup_main_thread",__FILE__,__LINE__);
}




typedef struct last_pos {
	HTID		tid;
	char * 		file;
	int		line;
} LAST_POS;

LAST_POS pp;

void
__u_cp(char * pos,int line);

void
__u_cp(char * pos,int line)
{
extern D_SEM task_lock;
	_lock_task(task_lock,__FILE__,__LINE__);
	pp.tid = _get_tid();
	pp.file = pos;
	pp.line = line;
	_unlock_task(task_lock,"u_cp",__FILE__,__LINE__);
}

void
send_alarm(int tid)
{
	pthread_kill((pthread_t)get_hard_tid(tid).tid, SIGALRM);
}


typedef struct arg_t {
	int 		(*main_ptr)(int,char**);
	int 		argc;
	char ** 	argv;
	int 		ret;
	unsigned	exit_flag:1;
	int		children_nos;
} ARG_T;

ARG_T call_main_arg;

int
u_child_countup()
{
int ret;
int f;
	f = 0;
	_lock_task(task_lock,__FILE__,__LINE__);
	if ( call_main_arg.exit_flag ) {
		ret = -1;
		f = 1;
	}
	else {
		if ( call_main_arg.children_nos == 0 )
			call_main_arg.children_nos ++;
		ret = 0;
	}
	_unlock_task(task_lock,"call_main",__FILE__,__LINE__);
	if ( f )
		for ( ; ; )
			sleep_sec(1);
	return ret;
}

void
u_child_countdown()
{
	_lock_task(task_lock,__FILE__,__LINE__);
	if ( call_main_arg.children_nos > 0 ) {
		call_main_arg.children_nos --;
	}
	_unlock_task(task_lock,"call_main",__FILE__,__LINE__);
}

void
u_exit(int val)
{
	_lock_task(task_lock,__FILE__,__LINE__);
/*
	if ( call_main_arg.children_nos ) {
		exit(val);
	}
	else {
*/
		call_main_arg.ret = val;
		call_main_arg.exit_flag = 1;
		t_wakeup_task((int)&call_main_arg,0);
/*
      	}
*/

	_unlock_task(task_lock,"call_main",__FILE__,__LINE__);

	exit(val);
}

void
call_main_task()
{
int ret;
	ret = (*call_main_arg.main_ptr)(
			call_main_arg.argc,
			call_main_arg.argv);
	u_exit(ret);
}


void
wait_thread()
{
int er;
	for ( ; /* call_main_arg.exit_flag == 0 */ ; ) {
		er = wait(0);
		if ( er < 0 && errno == ECHILD ) {
		  sleep_sec(2);
		}
	}
}

int
call_main(int argc,char ** argv,int (*main_ptr)(int,char**))
{

	call_main_arg.main_ptr = main_ptr;
	call_main_arg.argc = argc;
	call_main_arg.argv = argv;
	call_main_arg.ret = 0;

	if ( ignore_child_flag  )
		create_task(wait_thread,0,PRI_MAX-1);

	call_main_task();
//	create_task(call_main_task,0,PRI_MAX-1);

	_lock_task(task_lock,__FILE__,__LINE__);
	for ( ; call_main_arg.exit_flag == 0 ; ) {
		t_sleep_task((int)&call_main_arg,task_lock,0);

		_lock_task(task_lock,__FILE__,__LINE__);
	}	
	_unlock_task(task_lock,"call_main",__FILE__,__LINE__);



	ufr_close_all();

	if ( ignore_child_flag ) {
		for ( ; ; ) {
			if ( wait(0) < 0 )
				break;
		}
	}
	return call_main_arg.ret;
}

