/**********************************************************************
 
	Copyright (C) 2003-2007 Hirohisa MORI <joshua@globalbase.org>
 
	This program is free software; you can redistribute it 
	and/or modify it under the terms of the GLOBALBASE 
	Library General Public License (G-LGPL) as published by 

	http://www.globalbase.org/
 
	This program is distributed in the hope that it will be 
	useful, but WITHOUT ANY WARRANTY; without even the 
	implied warranty of MERCHANTABILITY or FITNESS FOR A 
	PARTICULAR PURPOSE.

**********************************************************************/


#include	<stdlib.h>
#include	<fcntl.h>
#include	<math.h>
#include	"memory_debug.h"
#include	"xl.h"
#include	"gbparam.h"
#include	"utils.h"
#include	"gbview.h"
#include	"memory_routine.h"
#include	"xlerror.h"
#include	"tree_cache.h"
#include	"lock_level.h"
#include	"pri_level.h"
#include	"win_flame.h"
#include	"machine/include.h"
#include	"mx_format.h"
#include	"pg_vector.h"
#include	"pg_scale.h"

typedef struct pixel_req {
	struct pixel_req *	next;
	int			loop_no;
	WF_ID	 		wfid;
	RESOURCE *		r;
	int			level;
	int			x;
	int			y;
} PIXEL_REQ;

extern int spiral_count;
void pixel_req_task();
void pixel_gc_task();

PIXEL_REQ * delete_pixel_req();
void clean_prq(PIXEL_REQ * pr);
void purge_prq(PIXEL_REQ * pr);
void set_pixel_dirty(PIXEL_REQ * pr);

extern SEM luster_lock,tc_lock;
extern XLISP_ENV * gv_resource_env[];
extern XLISP_ENV * luster_cmd;
PIXEL_REQ *	pixel_req_queue;
PIXEL_REQ * 	prq_free;
int queue_length;
void pixel_gc_tick();
int get_pixel_r64();
int get_pixel_cr();
int get_query_r64();
int get_query_cr();
int exit_lock_luster();
void
draw_pixel_data_2(
	VIEW_METHOD * vm,
	GBVIEW_FLAME * gf,
	RESOURCE* r,
	DRAW_WORK * dw);
int
vmt_create_draw_pixel(VIEW_METHOD*,MATRIX_NODE *);
void
vmt_dummy(VIEW_METHOD* vm,GBVIEW_FLAME* gf,RESOURCE * r,DRAW_WORK* dw);

extern VIEW_METHOD_TBL vm_plot_tbl;

VIEW_METHOD_TBL vm_rgb_tbl = {
	MDT_RGB8|MDT_VECTOR,0,
	vmt_create_draw_pixel,
	0,
	draw_pixel_data_2,
	0
};

VIEW_METHOD_TBL * pixel_vm[] = {
	&vm_rgb_tbl,
	&vm_plot_tbl,
	&vm_scale_tbl_int8,
	&vm_scale_tbl_uint8,
	&vm_scale_tbl_int16,
	&vm_scale_tbl_uint16,
	&vm_scale_tbl_int32,
	&vm_scale_tbl_uint32,
	&vm_scale_tbl_int64,
	&vm_scale_tbl_uint64,
	&vm_scale_tbl_float,
	&vm_scale_tbl_double,
	0
};

void
init_luster_r64()
{
XLISP_ENV * l_top;
XL_SEXP * gv_set_visible_resolution();
	init_luster();
	vm_tbl[RT_PIXEL_MAP_R64] = pixel_vm;
	l_top = gv_resource_env[RT_PIXEL_MAP_R64];
	if ( l_top == 0 )
		goto next;
	set_env(l_top,l_string(std_cm,"set-visible-resolution"),
			get_func_prim(gv_set_visible_resolution,
			FO_APPLICATIVE,0,2,2));
	init_pg_vector(luster_cmd);
	init_pg_scale(luster_cmd);
next:
/*
	create_task(pixel_req_task,0,PRI_FETCH);
*/
	new_tick((void(*)(int))pixel_gc_tick,1,0);
	
	init_old_matrix();
}
/*int *debug_chace=0;*/

#ifdef USE_MATRIX
void
gv_new_luster_option(RESOURCE * r,int r64_level)
{
ss_printf("new_luster_option\n");
	switch ( r->pr64.type ) {
	case 'P':
	case 'G':
	case 'B':
		r->h.mtx = open_matrix(get_url_str2(&r->h.target),0,0,
						open_old_matrix,get_omx_r64_work(r));
		break;
	case '0':
	case '1':
	case '2':
		r->h.mtx = open_matrix(get_url_str2(&r->h.target),0,0,
						open_old_matrix,get_omx_cr_work(r));
		break;
	case 'M':
		insert_resource_matrix(r,1);
		break;
	default:
		er_panic("gv_new_luster_option(1)");
	}
}

#else /* USE_MATRIX */

void
gv_new_luster_option(RESOURCE * r,int r64_level)
{
int ses;
	ses = open_session(SEST_OPTIMIZE);
	switch ( r->pr64.type ) {
	case 'P':
	case 'G':
	case 'B':
		init_tc(ses,&r->pr64.tc,r64_level,RECT_LOGSIZE,
			get_pixel_r64,get_query_r64,r);
		break;
	case '0':
	case '1':
	case '2':
		init_tc(ses,&r->pr64.tc,r64_level,RECT_LOGSIZE,
			get_pixel_cr,get_query_cr,r);
/*		
		if(debug_chace==0){
			debug_chace = &(r->pr64.tc.cache);
		}
*/		break;
	default:
		er_panic("gv_new_luster_option(1)");
	}
	close_session(ses);
}

#endif /* USE_MATRIX */

void
insert_pixel_req(WF_ID wfid,RESOURCE * r,int lev,int x,int y)
{
PIXEL_REQ * pr;
	if ( prq_free ) {
		pr = prq_free;
		prq_free = pr->next;
	}
	else	pr = d_alloc(sizeof(*pr));
	pr->loop_no = r->h.loop_no;
	pr->wfid = wfid;
	pr->r = r;
	pr->x = x;
	pr->y = y;
	pr->level = lev;
	pr->next = pixel_req_queue;
	pixel_req_queue = pr;
	queue_length ++;
}

PIXEL_REQ *
delete_pixel_req()
{
PIXEL_REQ * pr;
	pr = pixel_req_queue;
	pixel_req_queue = pr->next;
	queue_length --;
	return pr;
}

void
clean_prq(PIXEL_REQ * pr)
{
PIXEL_REQ * pr1,** prp;
int _yy,_yy1;
int _xx,_xx1;
	for ( prp = &pixel_req_queue ; *prp ; ) {
		pr1 = *prp;
		if ( pr1->r != pr->r )
			goto next;
		if ( pr1->wfid != pr->wfid )
			goto next;
		if ( pr1->level < pr->level )
			goto next;
		_xx1 = pr1->x&(-(RECT_SIZE<<pr1->level));
		_xx = pr->x&(-(RECT_SIZE<<pr1->level));
		if ( _xx != _xx1 )
			goto next;
		_yy1 = pr1->y&(-(RECT_SIZE<<pr1->level));
		_yy = pr->y&(-(RECT_SIZE<<pr1->level));
		if ( _yy != _yy1 )
			goto next;
		*prp = pr1->next;
		pr1->next = prq_free;
		prq_free = pr1;
		queue_length --;
		continue;
	next:
		prp = &pr1->next;
	}
	pr->next = prq_free;
	prq_free = pr;
}

void
purge_prq(PIXEL_REQ * pr)
{
RESOURCE * r;
PIXEL_REQ * pr1, ** prp;
	r = pr->r;
	for ( prp = &pixel_req_queue ; *prp ; ) {
		pr1 = *prp;
		if ( pr1->r == r ) {
			*prp = pr1->next;
			pr1->next = prq_free;
			prq_free = pr1;
			queue_length --;
		}
		else {
			prp = &pr1->next;
		}
	}
	pr->next = prq_free;
	prq_free = pr;
}

void
pixel_gc_tick()
{
int i;
PIXEL_REQ * pr;
	lock_task(luster_lock);
	for ( i = 0 ; i < 20000 && prq_free ; i ++ ) {
		pr = prq_free;
		prq_free = pr->next;
		d_f_ree(pr);
	}
	unlock_task(luster_lock,"tick");
}


void
set_pixel_dirty(PIXEL_REQ * pr)
{
int length;
GB_RECT r;
int pic;
int x,y;
int mask;
	length = RECT_SIZE<<pr->level;
	mask = -length;
	x = pr->x & mask;
	y = pr->y & mask;
	r.tl.x = x / pr->r->pr64.dpm;
	r.tl.y = y / pr->r->pr64.dpm;
	pic = x + length;
	if ( pic > pr->r->pr64.width[0] )
		pic = pr->r->pr64.width[0];
	r.br.x = pic / pr->r->pr64.dpm;
	pic = pr->y + length;
	if ( pic > pr->r->pr64.height[0] )
		pic = pr->r->pr64.height[0];
	r.br.y = pic / pr->r->pr64.dpm;
	wf_insert_dirty_rect(pr->r,&r,WFF_LUSTER_DIRTY,120,10);
}

/*
void
pixel_req_task()
{
int lev2;
int ses;
XL_INTERPRETER * xli;
PIXEL_REQ * pr;
int i;
	xli = new_xl_interpreter();
	xli->a_type = XLA_SELF;
	setup_i(xli);
	ses = open_session(SEST_OPTIMIZE);


	for ( ; ; ) {

		lock_task(luster_lock);
		for ( ; pixel_req_queue == 0 ; ) {
			sleep_task((int)&pixel_req_queue,luster_lock);
			lock_task(luster_lock);
		}
		pr = delete_pixel_req();
		if ( pr->loop_no < pr->r->h.loop_no - spiral_count - 2 ) {
			purge_prq(pr);
			unlock_task(luster_lock,"pixel_req_task");
			continue;
		}
		unlock_task(luster_lock,"pixel_req_task");

		tc_get_pixel(ses,&pr->r->pr64.tc,
			pr->level,
			pr->x,
			pr->y,1,
			&lev2);
		if ( lev2 == pr->level ) {
			set_pixel_dirty(pr);
		}
		lock_task(luster_lock);
		clean_prq(pr);
		unlock_task(luster_lock,"pixel_req_task");
	}
}
*/

#ifdef USE_MATRIX

typedef struct mtx_cache {
	int		tl_x;
	int		tl_y;
	int		br_x;
	int		br_y;
	int		level;
	int		rect_logsize;
	MATRIX_DH_SET	ds;
	MATRIX_NODE *	n;
	unsigned long *	ds_ptr;
	unsigned	leaf;
} MTX_CACHE;

unsigned long
get_mtx_data(MTX_CACHE * cache,RESOURCE * r,int lev,int x,int y,int * lev_p);

unsigned long
get_mtx_data(MTX_CACHE * cache,RESOURCE * r,int lev,int x,int y,int * lev_p)
{
int err;
void * d;
MATRIX * m;
MATRIX_NODE * nn;
INTEGER64 dim_code[3];
int xx,yy;
INTEGER64 mx,my;
int clr;
	m = r->h.mtx;
	if ( m == 0 ) {
		gv_new_luster_option(r,0);
		if ( r->h.mtx == 0 )
			return 0;
	}
	if ( m->flags & MXF_ERR_DETECTED ) {
			r->h.mtx = 0;
		close_matrix(m);
		gv_new_luster_option(r,0);
		if ( r->h.mtx == 0 )
			return 0;
	}
	clr = 0;
	if ( cache->n ) {
		if ( cache->level == lev &&
			cache->tl_x <= x && cache->tl_y <= y &&
			cache->br_x > x && cache->br_y > y )
			goto end;
		clr = 1;
		matrix_node_channel_unlock(cache->n,0);
		unlock_node(cache->n,0);
	}
	nn = cache->n;
	
	dim_code[0] = lev;
	dim_code[1] = x;
	dim_code[2] = y;
//ss_printf("get_mtx_data-1\n");

	d = get_matrix_node_channel
			(&err,&cache->n,r->h.mtx,
			&dim_code[0],CH_VISU,
			GN_TREE,GN_NODE_CREATE,0,r->h.loop_no);
	if ( d == 0 ) {
		if ( nn && clr == 0 ) {
			matrix_node_channel_unlock(nn,0);
			unlock_node(nn,0);
		}
		memset(cache,0,sizeof(*cache));
		*lev_p = m->total_levels;
		return C_TRANSPARENT;
	}
	get_matrix_dh_set(&cache->ds,d);
	dim_code[0] = cache->n->dim_code[0];
	dim_code[1] = cache->n->dim_code[1];
	dim_code[2] = cache->n->dim_code[2];
	cache->tl_x = dim_code[1] & (mx=(-(1<<(dim_code[0]*m->dim_divide[0] + m->block_size[0]))));
	cache->tl_y = dim_code[2] & (my=(-(1<<(dim_code[0]*m->dim_divide[1] + m->block_size[1]))));
	cache->br_x = cache->tl_x + (1<<(dim_code[0]*m->dim_divide[0] + m->block_size[0]));
	cache->br_y = cache->tl_y + (1<<(dim_code[0]*m->dim_divide[1] + m->block_size[1]));
if ( (x & mx) != cache->tl_x || (y & my) != cache->tl_y )
er_panic("get_mtx_data");
	if ( cache->br_x > m->pixel_size[0] )
		cache->br_x = m->pixel_size[0];
	if ( cache->br_y > m->pixel_size[1] )
		cache->br_y = m->pixel_size[1];
	cache->level = dim_code[0];
	cache->ds_ptr = cache->ds.offset;
	cache->rect_logsize = m->block_size[0];
	*lev_p = dim_code[0];

end:
	xx = (x - cache->tl_x)>>(cache->level*m->dim_divide[0]);
	yy = (y - cache->tl_y)>>(cache->level*m->dim_divide[1]);
	return cache->ds_ptr[xx + yy * cache->ds.ix[0]];
}


unsigned long
convert_rgba_to_cc(unsigned int cc)
{
unsigned int r,g,b;
	if ( (cc & 0xff) >= 0x80 )
		return C_TRANSPARENT;
	r = (cc >> 24)&0x0ff;
	g = (cc >> 16)&0x0ff;
	b = (cc >> 8)&0x0ff;
	return ((r << (COL_BIT-8))) |
		((g << (COL_BIT-8)) << COL_BIT) |
		((b << (COL_BIT-8)) << (COL_BIT*2));
}



unsigned long
get_mtx_data_2(VIEW_METHOD * vm,MTX_CACHE * cache,RESOURCE * r,INTEGER64 lev,INTEGER64 x,INTEGER64 y,INTEGER64 * lev_p);

unsigned long
get_mtx_data_2(VIEW_METHOD * vm,MTX_CACHE * cache,RESOURCE * r,INTEGER64 lev,INTEGER64 x,INTEGER64 y,INTEGER64 * lev_p)
{
int err;
void * d;
MATRIX * m;
MATRIX_NODE * nn;
INTEGER64 dim_code[3];
int xx,yy;
INTEGER64 mx,my;
int clr;
	m = r->h.mtx;
	clr = 0;
	if ( cache->n ) {
		if ( cache->level == lev &&
			cache->tl_x <= x && cache->tl_y <= y &&
			cache->br_x > x && cache->br_y > y )
			goto end;
		clr = 1;
		matrix_node_channel_unlock(cache->n,0);
		unlock_node(cache->n,0);
	}
	nn = cache->n;
	
	dim_code[0] = lev;
	dim_code[1] = x;
	dim_code[2] = y;
//ss_printf("get_mtx_data-1\n");

	if ( r->h.mtx->p.flags & MPF_NODE_DIPENDENT ) {
		d = get_matrix_node_channel
				(&err,&cache->n,r->h.mtx,
				&dim_code[0],vm->view_channel[0],
				GN_TREE,GN_NODE_CREATE,0,r->h.loop_no);
	}
	else {
		d = get_matrix_node_channel
				(&err,&cache->n,r->h.mtx,
				&dim_code[0],vm->view_channel[0],
				GN_TREE,GN_STEP_NODE_CREATE,0,r->h.loop_no);
	}
	if ( d == 0 ) {
		if ( nn && clr == 0 ) {
			matrix_node_channel_unlock(nn,0);
			unlock_node(nn,0);
		}
		memset(cache,0,sizeof(*cache));
		*lev_p = m->total_levels;
		return C_TRANSPARENT;
	}
	get_matrix_dh_set(&cache->ds,d);
	dim_code[0] = cache->n->dim_code[0];
	dim_code[1] = cache->n->dim_code[1];
	dim_code[2] = cache->n->dim_code[2];
	cache->tl_x = dim_code[1] & (mx=(-(1<<(dim_code[0]*m->dim_divide[0] + m->block_size[0]))));
	cache->tl_y = dim_code[2] & (my=(-(1<<(dim_code[0]*m->dim_divide[1] + m->block_size[1]))));
	cache->br_x = cache->tl_x + (1<<(dim_code[0]*m->dim_divide[0] + m->block_size[0]));
	cache->br_y = cache->tl_y + (1<<(dim_code[0]*m->dim_divide[1] + m->block_size[1]));
if ( (x & mx) != cache->tl_x || (y & my) != cache->tl_y )
er_panic("get_mtx_data");
	if ( cache->br_x > m->pixel_size[0] )
		cache->br_x = m->pixel_size[0];
	if ( cache->br_y > m->pixel_size[1] )
		cache->br_y = m->pixel_size[1];
	cache->level = dim_code[0];
	cache->ds_ptr = cache->ds.offset;
	cache->rect_logsize = m->block_size[0];
	*lev_p = dim_code[0];

end:
	xx = (x - cache->tl_x)>>(cache->level*m->dim_divide[0]);
	yy = (y - cache->tl_y)>>(cache->level*m->dim_divide[1]);
	return convert_rgba_to_cc(cache->ds_ptr[xx + yy * cache->ds.ix[0]]);
}


void
draw_pixel_data_1(
	GBVIEW_FLAME * gf,
	RESOURCE* r,
	DRAW_WORK * dw)
{
int i;
GB_POINT p;
REAL1 res;
int xx,yy;
REAL1 dpm1,dpm2;
int level;
float d1,d2,d;
unsigned long cc;
int lev;
int lev_cnt;
MTX_CACHE cache;
int tim,tim2;

	if ( r->pr64.exit_lock )
		return;
	if ( get_mov_flag(gf,dw->wfid) != 0 )
		return;

	dw->limit_reso = r->pr64.dpm;
	dw->limit_ptr = dw->pt_list[dw->size/2];
	r->h.loop_no ++;
	dw->flags |= WFF_LUSTER; 
	dpm1 = r->pr64.dpm;
	dpm2 = dpm1;
	level = 0;
	d1 = 1;
	d2 = 0;

	res = -1;
	for ( i = 0 ; i < dw->size ; i ++ ) {
		if ( res < dw->pt_reso[i] )
			res = dw->pt_reso[i];
	}

/*
	if ( res > dpm1*LIMIT_RESOLUTION_RATE )
		return;
*/

	if ( dpm2 < res || res < dpm2/2 ) {
		if ( res > dpm1 ) {
			dpm2 = dpm1;
		level = 0;
		}
		else {
			level = 0;
			d = -1;
			for ( dpm2 = dpm1 ;
				res <= dpm2 &&
				level < r->pr64.max_level;
				d = dpm2,
				dpm2 /= 2,
				level ++ );
			dpm2 = d;
			level --;
		}
	}

	lock_task(luster_lock);
//	lock_task(tc_lock);
	lev_cnt = 0;
	memset(&cache,0,sizeof(cache));
	lev = level;
	tim = get_xltime();
	for ( i = 0 ; i < dw->size ; i ++ ) {
		if ( (i % 1000) == 0 && wf_active_check_2(gf,dw->wfid) < 0 )
{ss_printf("FREE=RETURN-PIXEL\n");
			break;
}
		if ( r->pr64.exit_lock )
			break;
		tim2 = get_xltime();
		if ( tim != tim2 ) {
			if ( get_mov_flag(gf,dw->wfid) != 0 )
				break;
			tim = tim2;
		}
		p = dw->pt_list[i];
		if ( dw->pt_reso[i] < 0 )
			continue;
		xx = rint(p.x*r->pr64.dpm);
		yy = rint(p.y*r->pr64.dpm);
		if ( xx < 0 || yy < 0 )
			continue;
		if ( xx >= r->pr64.width[0] )
			continue;
		if ( yy >= r->pr64.height[0] )
			continue;
		if ( lev_cnt && cache.ds_ptr ) {
		int lx,ly;
			if (	cache.tl_x <= xx && xx < cache.br_x &&
				cache.tl_y <= yy && yy < cache.br_y &&
				cache.level == lev ) {

				lx = (xx>>(lev))&((1<<cache.rect_logsize)-1);
				ly = (yy>>(lev))&((1<<cache.rect_logsize)-1);
				if ( lx >= cache.ds.ix[0] ) {
					cc = C_TRANSPARENT;
				}
				else if ( ly >= cache.ds.ix[1] ) {
					cc = C_TRANSPARENT;
				}
				else cc = cache.ds_ptr[lx + 
					ly*cache.ds.ix[0]];
			}
			else goto non_cache;
			lev_cnt ++;
			if ( lev_cnt >= 100 )
				lev_cnt = 0;
		}
		else {
		non_cache:
//			cc = _tc_get_pixel(&r->pr64.tc,level,xx,yy,&lev);
			cc = get_mtx_data(&cache,r,level,xx,yy,&lev);
			if ( level != lev )
				dw->data_request ++;
//			cache = r->pr64.tc;
			lev_cnt ++;
		}
		dw->pixels[i] = cc;
		dw->flags |= WFF_DRAW;
	}
//	unlock_task(tc_lock,"draw_pixel");

	if ( cache.n ) {
		matrix_node_channel_unlock(cache.n,0);
		unlock_node(cache.n,0);
	}
	

	unlock_task(luster_lock,"draw_pixel");

}


void
draw_pixel_data_2(
	VIEW_METHOD * vm,
	GBVIEW_FLAME * gf,
	RESOURCE* r,
	DRAW_WORK * dw)
{
int i;
GB_POINT p;
REAL1 res;
INTEGER64 xx,yy;
REAL1 dpm1,dpm2;
INTEGER64 level;
REAL1 d;
unsigned long cc;
INTEGER64 lev;
INTEGER64 lev_cnt;
MTX_CACHE cache;
INTEGER64 tim,tim2;

MATRIX * mtx;
REAL1 divide;
INTEGER64 int_divide;


	if ( r->pr64.exit_lock )
		return;
	if ( get_mov_flag(gf,dw->wfid) != 0 )
		return;
	if ( r->h.mtx == 0 ) {
		gv_new_luster_option(r,0);
		if ( r->h.mtx == 0 )
			return;
	}
	mtx = r->h.mtx;
	if ( mtx->flags & MXF_ERR_DETECTED ) {
		r->h.mtx = 0;
		close_matrix(mtx);
		gv_new_luster_option(r,0);
		if ( r->h.mtx == 0 )
			return;
		mtx = r->h.mtx;
	}

	divide = 1;
	for ( i = 0 ; i < mtx->dim_divide[0] ; i ++ )
		divide *= 2;
	int_divide = mtx->dim_divide[0];

	dw->limit_reso = r->pr64.dpm;
	dw->limit_ptr = dw->pt_list[dw->size/2];
	r->h.loop_no ++;
	dw->flags |= WFF_LUSTER; 
	dpm1 = r->pr64.dpm;
	dpm2 = dpm1;
	level = 0;

	res = -1;
	for ( i = 0 ; i < dw->size ; i ++ ) {
		if ( res < dw->pt_reso[i] )
			res = dw->pt_reso[i];
	}

/*
	if ( res > dpm1*LIMIT_RESOLUTION_RATE )
		return;
*/

	if ( dpm2 < res || res < dpm2/divide ) {
		if ( res > dpm1 ) {
			dpm2 = dpm1;
		level = 0;
		}
		else {
			level = 0;
			d = -1;
			for ( dpm2 = dpm1 ;
				res <= dpm2 &&
				level < r->pr64.max_level;
				d = dpm2,
				dpm2 /= divide,
				level ++ );
			dpm2 = d;
			level --;
		}
	}

	lock_task(luster_lock);
//	lock_task(tc_lock);
	lev_cnt = 0;
	memset(&cache,0,sizeof(cache));
	lev = level;
	tim = get_xltime();
	for ( i = 0 ; i < dw->size ; i ++ ) {
		if ( (i % 1000) == 0 && wf_active_check_2(gf,dw->wfid) < 0 )
{ss_printf("FREE=RETURN-PIXEL\n");
			break;
}
		if ( r->pr64.exit_lock )
			break;
		tim2 = get_xltime();
		if ( tim != tim2 ) {
			if ( get_mov_flag(gf,dw->wfid) != 0 )
				break;
			tim = tim2;
		}
		p = dw->pt_list[i];
		if ( dw->pt_reso[i] < 0 )
			continue;
		xx = rint(p.x*r->pr64.dpm);
		yy = rint(p.y*r->pr64.dpm);
		if ( xx < 0 || yy < 0 )
			continue;
		if ( xx >= mtx->pixel_size[0] )
			continue;
		if ( yy >= mtx->pixel_size[1] )
			continue;
		if ( lev_cnt && cache.ds_ptr ) {
		int lx,ly;
			if (		cache.tl_x <= xx && xx < cache.br_x &&
					cache.tl_y <= yy && yy < cache.br_y ) {
				if ( cache.level == lev ) {

					lx = (xx>>(lev*int_divide))&((((INTEGER64)1)<<mtx->block_size[0])-1);
					ly = (yy>>(lev*int_divide))&((((INTEGER64)1)<<mtx->block_size[1])-1);
					if ( lx >= cache.ds.ix[0] ) {
						cc = C_TRANSPARENT;
					}
					else if ( ly >= cache.ds.ix[1] ) {
						cc = C_TRANSPARENT;
					}
					else cc = convert_rgba_to_cc(cache.ds_ptr[lx + 
						ly*cache.ds.ix[0]]);
				}
				else if ( cache.leaf ) {
					lx = (xx>>(cache.level*int_divide))&((((INTEGER64)1)<<mtx->block_size[0])-1);
					ly = (yy>>(cache.level*int_divide))&((((INTEGER64)1)<<mtx->block_size[1])-1);
					if ( lx >= cache.ds.ix[0] ) {
						cc = C_TRANSPARENT;
					}
					else if ( ly >= cache.ds.ix[1] ) {
						cc = C_TRANSPARENT;
					}
					else cc = convert_rgba_to_cc(cache.ds_ptr[lx + 
						ly*cache.ds.ix[0]]);
				}
				else	goto non_cache;
			}
			else goto non_cache;
			lev_cnt ++;
			if ( lev_cnt >= 100 )
				lev_cnt = 0;
		}
		else {
		non_cache:
//			cc = _tc_get_pixel(&r->pr64.tc,level,xx,yy,&lev);
			cc = get_mtx_data_2(vm,&cache,r,level,xx,yy,&lev);
//ss_printf("CC %lli %lli %lli %x\n",xx,yy,level,cc);
			if ( level != lev && cache.leaf == 0 )
				dw->data_request ++;
//			cache = r->pr64.tc;
			lev_cnt ++;
		}
		dw->pixels[i] = cc;
		dw->flags |= WFF_DRAW;
	}
//	unlock_task(tc_lock,"draw_pixel");

	if ( cache.n ) {
		matrix_node_channel_unlock(cache.n,0);
		unlock_node(cache.n,0);
	}
	
	unlock_task(luster_lock,"draw_pixel");

}

int
vmt_create_draw_pixel(VIEW_METHOD* vm,MATRIX_NODE *n)
{
	vm->work = 0;
	return 0;
}



void
draw_pixel_data(
	GBVIEW_FLAME * gf,
	RESOURCE* r,
	DRAW_WORK * dw)
{
VIEW_METHOD * vm;
	switch ( r->pr64.type ) {
	case 'M':
//		draw_pixel_data_2(gf,r,dw);
		if ( r->h.vm == 0 ) {
			vmt_dummy(0,gf,r,dw);
		}
		else {
			for ( vm = r->h.vm ; vm ; vm = vm->next )
				(*vm->tbl->vmt_draw)(vm,gf,r,dw);
		}
		break;
	default:
		draw_pixel_data_1(gf,r,dw);
		break;
	}
}

#else
void
draw_pixel_data(
	GBVIEW_FLAME * gf,
	RESOURCE* r,
	DRAW_WORK * dw)
{
int i;
GB_POINT p;
REAL1 res;
int xx,yy;
REAL1 dpm1,dpm2;
int level;
float d1,d2,d;
unsigned long cc;
int lev;
int lev_cnt;
TREE_CACHE cache;
int tim,tim2;

	if ( r->pr64.exit_lock )
		return;
	if ( get_mov_flag(gf,dw->wfid) != 0 )
		return;

	dw->limit_reso = r->pr64.dpm;
	dw->limit_ptr = dw->pt_list[dw->size/2];
	r->h.loop_no ++;
	dw->flags |= WFF_LUSTER; 
	dpm1 = r->pr64.dpm;
	dpm2 = dpm1;
	level = 0;
	d1 = 1;
	d2 = 0;

	res = -1;
	for ( i = 0 ; i < dw->size ; i ++ ) {
		if ( res < dw->pt_reso[i] )
			res = dw->pt_reso[i];
	}

/*
	if ( res > dpm1*LIMIT_RESOLUTION_RATE )
		return;
*/
	if ( dpm2 < res || res < dpm2/2 ) {
		if ( res > dpm1 ) {
			dpm2 = dpm1;
		level = 0;
		}
		else {
			level = 0;
			d = -1;
			for ( dpm2 = dpm1 ;
				res <= dpm2 &&
				level < r->pr64.max_level;
				d = dpm2,
				dpm2 /= 2,
				level ++ );
			dpm2 = d;
			level --;
		}
	}

	lock_task(luster_lock);
	lock_task(tc_lock);
	lev_cnt = 0;
	cache.cache = 0;
	lev = level;
	tim = get_xltime();
	for ( i = 0 ; i < dw->size ; i ++ ) {
		if ( r->pr64.exit_lock )
			break;
		tim2 = get_xltime();
		if ( tim != tim2 ) {
			if ( get_mov_flag(gf,dw->wfid) != 0 )
				break;
			tim = tim2;
		}
		p = dw->pt_list[i];
		if ( dw->pt_reso[i] < 0 )
			continue;
		xx = rint(p.x*r->pr64.dpm);
		yy = rint(p.y*r->pr64.dpm);
		if ( xx < 0 || yy < 0 )
			continue;
		if ( xx >= r->pr64.width[0] )
			continue;
		if ( yy >= r->pr64.height[0] )
			continue;
		if ( lev_cnt && cache.cache ) {
		int lx,ly;
			if (	cache.tl_x <= xx && xx < cache.br_x &&
				cache.tl_y <= yy && yy < cache.br_y &&
				cache.cache->level == lev ) {

				lx = (xx>>(lev))&((1<<cache.rect_logsize)-1);
				ly = (yy>>(lev))&((1<<cache.rect_logsize)-1);
				if ( lx >= cache.cache->w ) {
					cc = C_TRANSPARENT;
				}
				else if ( ly >= cache.cache->h ) {
					cc = C_TRANSPARENT;
				}
				else cc = cache.cache->pixels[lx + 
					ly*cache.cache->w];
			}
			else goto non_cache;
			lev_cnt ++;
			if ( lev_cnt >= 100 )
				lev_cnt = 0;
		}
		else {
		non_cache:
			cc = _tc_get_pixel(&r->pr64.tc,level,xx,yy,&lev);
			if ( level != lev )
				dw->data_request ++;
			cache = r->pr64.tc;
			lev_cnt ++;
		}
		dw->pixels[i] = cc;
		dw->flags |= WFF_DRAW;
	}
	unlock_task(tc_lock,"draw_pixel");
	unlock_task(luster_lock,"draw_pixel");

}

#endif

int
exit_lock_luster(RESOURCE * r,int data)
{
	r->pr64.exit_lock = data;
	return 0;
}


int
get_point_pixel_data(
	GBVIEW_FLAME * gf,
	RESOURCE *	r,
	GET_POINT_WORK * w)
{
INDICATE * in;
GB_POINT p[2]; 
	memcpy(p,w->pt,sizeof(p));
	if ( p[0].x < 0 || p[0].y < 0 )
		return -1;
	if ( p[0].x >= r->pr64.width[0]/r->pr64.dpm ) {
		return -1;
	}
	if ( p[0].y >= r->pr64.height[0]/r->pr64.dpm ) {
		return -1;
	}
	in = _insert_indicate(&w->indicate_list,&r->h.entry);
	memcpy(in->result,p,sizeof(p));
	in->unit = ll_copy_str(r->h.cu.unit);
	in->status = IS_COORDINATE;
	return 0;
}


