/**********************************************************************
 
	Copyright (C) 2003 Hirohisa MORI <joshua@nichibun.ac.jp>
 
	This program is free software; you can redistribute it 
	and/or modify it under the terms of the GLOBALBASE 
	Library General Public License (G-LGPL) as published by 

	http://www.globalbase.org/
 
	This program is distributed in the hope that it will be 
	useful, but WITHOUT ANY WARRANTY; without even the 
	implied warranty of MERCHANTABILITY or FITNESS FOR A 
	PARTICULAR PURPOSE.

**********************************************************************/

#include <stdlib.h>
#include <string.h>

#include "v/VgbFlame.h"
#include "v/vobj_utils.h"

extern "C" {
#include 	"xl.h"
#include	"xlerror.h"
#include	"radar.h"
#include	"memory_debug.h"

char *
VgbFlame_get_sw_from_sf(VgbFlameSwitchStatus * sts,int *flags,XL_SYM_FIELD * sf);
VObject * get_object_from_name(XLISP_ENV * env,L_CHAR * name);
char * VgbFlame_get_flame_from_sf(
	VgbFlameStatus * sts,
	int * flags,
	XLISP_ENV * env,
	XL_SYM_FIELD * sf,
	XL_SEXP * arg,
	VObject::FreeList **);
char * VgbFlame_get_flame_radar_from_sf(
	VgbFlameRadarStatus * sts,
	int * flags,
	XLISP_ENV * env,
	XL_SYM_FIELD * sf);
char * VgbFlame_get_flame_coordinate_from_sf(
	VgbFlameCoordinateStatus * sts,
	int * flags,
	XLISP_ENV * env,
	XL_SYM_FIELD * sf);
XL_SEXP * vobj_VgbFlameStandard(XLISP_ENV *env, XL_SEXP *arg, XLISP_ENV *a, XL_SYM_FIELD *sf);
void init_vobj_VgbFlameStandard(XLISP_ENV *env);
void free_VgbFlameRadarStatus(VgbFlameRadarStatus * sts,int flags);
VImage *
get_image_from_lchar(XL_SEXP ** retp,XLISP_ENV * env,L_CHAR *,L_CHAR * _img_field);
char * _get_point_from_char(REAL1*,char*);
GB_POINT get_point_from_char(L_CHAR*);
GB_RECT get_rect_from_char(L_CHAR*);
XL_SEXP *
vsat_VgbFlameStandard_translator(VObject* obj, void* arg);

XL_SEXP *
vobj_load_warp_point_menu(XLISP_ENV * env,XL_SEXP * s,XLISP_ENV * a,XL_SYM_FIELD * sf);
XL_SEXP *
gb_trackback(XLISP_ENV * env,XL_SEXP * s,XLISP_ENV * a,XL_SYM_FIELD * sf);

void
init_vobj_VgbFlameStandard(XLISP_ENV *env)
{
	set_env(env,l_string(std_cm,"VgbFlameStandard"),
		get_func_prim((XL_SEXP*(*)())vobj_VgbFlameStandard,FO_NORMAL,0,1,1));
	set_env(env,l_string(std_cm,"VgbFlameStandardLoadWPM"),
		get_func_prim((XL_SEXP*(*)())vobj_load_warp_point_menu,FO_APPLICATIVE,0,1,2));
	set_env(gblisp_top_env0,l_string(std_cm,"trackback"),
		get_func_prim((XL_SEXP*(*)())gb_trackback,FO_NORMAL,0,2,2));



	set_env(env,l_string(std_cm,"VMT-WARP"),
		n_get_string(VMT_WARP_STR));
	set_env(env,l_string(std_cm,"VMT-ADD-WARP"),
		n_get_string(VMT_ADD_WARP_STR));
	set_env(env,l_string(std_cm,"VMT-EDIT-WARP"),
		n_get_string(VMT_EDIT_WARP_STR));
	set_env(env,l_string(std_cm,"VMT-WARP-REPLACE-MODE"),
		n_get_string(VMT_WARP_REPLACE_MODE_STR));
	set_env(env,l_string(std_cm,"VMT-WARP-INSERT-MODE"),
		n_get_string(VMT_WARP_INSERT_MODE_STR));

	set_env(env,l_string(std_cm,"VMT-COPY-WARP-POINT"),
		n_get_string(VMT_COPY_WARP_POINT_STR));

	set_env(env,l_string(std_cm,"VMT-SET-IWP"),
		n_get_string(VMT_SET_IWP_STR));

	set_env(env,l_string(std_cm,"VMT-HWF"),
		n_get_string(VMT_HWF_STR));
}

char * _get_point_from_char(REAL1 * ret,char * str)
{
char * p;
char * d;
int len;
	for ( p = str ; *p && *p != ',' ; p ++ );
	d = (char*)d_alloc(len = p - str + 1);
	memcpy(d,str,len);
	d[p-str] = 0;
	sscanf(str,REAL1_F,ret);
	d_f_ree(d);
	if ( *p == 0 )
		return 0;
	return p+1;
}

GB_POINT get_point_from_char(L_CHAR * str)
{
GB_POINT ret;
char * d;
	d = n_string(std_cm,str);
	d = _get_point_from_char(&ret.x,d);
	if ( d )
		_get_point_from_char(&ret.y,d);
	return ret;
}

GB_RECT get_rect_from_char(L_CHAR * str)
{
GB_RECT ret;
char * d;
	d = n_string(std_cm,str);
	d = _get_point_from_char(&ret.tl.x,d);
	if ( d )
		d = _get_point_from_char(&ret.tl.y,d);
	if ( d )
		d = _get_point_from_char(&ret.br.x,d);
	if ( d )
		_get_point_from_char(&ret.br.y,d);
	return ret;
}



char * VgbFlame_get_flame_coordinate_from_sf(
	VgbFlameCoordinateStatus * sts,
	int * flags,
	XLISP_ENV * env,
	XL_SYM_FIELD * sf)
{
char * ret;
L_CHAR * d;
	*flags = 0;
	ret = 0;
	d = get_sf_attribute(sf,l_string(std_cm,"flame.crd.FlameBaseRect"));
	if ( d ) {
		sts->flame_base_rect = get_rect_from_char(d);
		(*flags) |= VSF_FLAME_C_FLAME_BASE_RECT;
	}
	d = get_sf_attribute(sf,l_string(std_cm,"flame.crd.FlameBaseCenter"));
	if ( d ) {
		sts->flame_base_center = get_point_from_char(d);
		(*flags) |= VSF_FLAME_C_FLAME_BASE_CENTER;
	}
	d = get_sf_attribute(sf,l_string(std_cm,"flame.crd.FlameBaseRotate"));
	if ( d ) {
		sscanf(n_string(std_cm,d),REAL1_F,&sts->flame_base_rotate);
		(*flags) |= VSF_FLAME_C_FLAME_BASE_ROTATE;
	}
	d = get_sf_attribute(sf,l_string(std_cm,"flame.crd.FlameBaseResolution"));
	if ( d ) {
		sscanf(n_string(std_cm,d),REAL1_F,&sts->flame_base_resolution);
		(*flags) |= VSF_FLAME_C_FLAME_BASE_RESOLUTION;
	}
	d = get_sf_attribute(sf,l_string(std_cm,"flame.crd.Target"));
	if ( d ) {
		sts->target = ll_copy_str(d);
		(*flags) |= VSF_FLAME_C_TARGET;
	}
	d = get_sf_attribute(sf,l_string(std_cm,"flame.crd.UseIWP"));
	if ( d ) {
		sts->use_iwp = atoi(n_string(std_cm,d));
		(*flags) |= VSF_FLAME_C_USE_IWP;
	}
	return 0;
}

char *
VgbFlame_get_sw_from_sf(VgbFlameSwitchStatus * sts,int *flags,XL_SYM_FIELD * sf)
{
char * ret;
L_CHAR * d;
	*flags = 0;
	ret = 0;
	d = get_sf_attribute(sf,l_string(std_cm,"flame.switch.Format"));
	if ( d ) {
		sts->format = ll_copy_str(d);
		(*flags) |= VSF_FLAME_S_FORMAT;
	}
	d = get_sf_attribute(sf,l_string(std_cm,"flame.switch.Quality"));
	if ( d ) {
		sts->quality = ll_copy_str(d);
		(*flags) |= VSF_FLAME_S_QUALITY;
	}
	d = get_sf_attribute(sf,l_string(std_cm,"flame.switch.Peripheral"));
	if ( d ) {
		sts->peripheral = ll_copy_str(d);
		(*flags) |= VSF_FLAME_S_PERIPHERAL;
	}
	d = get_sf_attribute(sf,l_string(std_cm,"flame.switch.Parameter"));
	if ( d ) {
		sts->parameter = ll_copy_str(d);
		(*flags) |= VSF_FLAME_S_PARAMETER;
	}
	d = get_sf_attribute(sf,l_string(std_cm,"flame.switch.AgentURL"));
	if ( d ) {
		sts->agent_url = ll_copy_str(d);
		(*flags) |= VSF_FLAME_S_AGENT_URL;
	}
	d = get_sf_attribute(sf,l_string(std_cm,"flame.switch.Mode"));
	if ( d ) {
		if ( l_strcmp(d,l_string(std_cm,"Idle")) == 0 )
			sts->mode = SW_MODE_IDLE;
		else if ( l_strcmp(d,l_string(std_cm,"Local")) == 0 )
			sts->mode = SW_MODE_LOCAL;
		else if ( l_strcmp(d,l_string(std_cm,"Remote")) == 0 )
			sts->mode = SW_MODE_REMOTE;
		else {
			ret = (char*)"flame.switch.Mode";
			goto err;
		}
		(*flags) |= VSF_FLAME_S_MODE;
	}
	return 0;
err:
	return ret;
}

XL_SEXP *
vsat_VgbFlameStandard_translator(VObject* obj, void* arg)
{
	return List(n_get_symbol("quote"),(XL_SEXP*)arg,-1);
}


char *
VgbFlame_get_flame_from_sf(
	VgbFlameStatus * sts,
	int * flags,
	XLISP_ENV * env,
	XL_SYM_FIELD * sf,
	XL_SEXP * arg,
	VObject::FreeList **free_list)
{
const char * ret;
XL_SEXP * _ret;
L_CHAR * d;
	*flags = 0;
	ret = 0;
	d = get_sf_attribute(sf,l_string(std_cm,"flame.Style"));
	if ( d ) {
		if ( l_strcmp(d,l_string(std_cm,"onFlameButton")) == 0 )
			sts->style = VGB_STYLE_ON_FLAME_BTN;
		else if ( l_strcmp(d,l_string(std_cm,"exFlameButton")) == 0 )
			sts->style = VGB_STYLE_EX_FLAME_BTN;
		else {
			ret = "flame.Style";
			goto err;
		}
		(*flags) |= VSF_FLAME_STYLE;
	}
	d = get_sf_attribute(sf,l_string(std_cm,"flame.DefaultWS"));
	if ( d ) {
		sts->default_ws = get_ws(d);
		if ( sts->default_ws == 0 ) {
			ret = "flame.DefaultWS";
			goto err;
		}
		(*flags) |= VSF_FLAME_DEFAULT_WS;
	}
	d = get_sf_attribute(sf,l_string(std_cm,"flame.ZoomIn"));
	if ( d )  {
		sts->zoom_in = (VButton*)get_object_from_name(env,d);
		if ( sts->zoom_in == 0 ) {
			ret = "flame.ZoomIn";
			goto err;
		}
		(*flags) |= VSF_FLAME_ZOOM_IN;
	}
	d = get_sf_attribute(sf,l_string(std_cm,"flame.ZoomOut"));
	if ( d )  {
		sts->zoom_out = (VButton*)get_object_from_name(env,d);
		if ( sts->zoom_out == 0 ) {
			ret = "flame.ZoomOut";
			goto err;
		}
		(*flags) |= VSF_FLAME_ZOOM_OUT;
	}
	d = get_sf_attribute(sf,l_string(std_cm,"flame.GrabModeButton"));
	if ( d )  {
		sts->grab_mode_button = (VRadioButton*)get_object_from_name(env,d);
		if ( sts->grab_mode_button == 0 ) {
			ret = "flame.GrabModeButton";
			goto err;
		}
		(*flags) |= VSF_FLAME_GRAB_MODE_BUTTON;
	}
	d = get_sf_attribute(sf,l_string(std_cm,"flame.RotateModeButton"));
	if ( d )  {
		sts->rotate_mode_button = (VRadioButton*)get_object_from_name(env,d);
		if ( sts->rotate_mode_button == 0 ) {
			ret = "flame.RotateModeButton";
			goto err;
		}
		(*flags) |= VSF_FLAME_ROTATE_MODE_BUTTON;
	}
	d = get_sf_attribute(sf,l_string(std_cm,"flame.ClickModeButton"));
	if ( d )  {
		sts->click_mode_button = (VRadioButton*)get_object_from_name(env,d);
		if ( sts->click_mode_button == 0 ) {
			ret = "flame.ClickModeButton";
			goto err;
		}
		(*flags) |= VSF_FLAME_CLICK_MODE_BUTTON;
	}
	d = get_sf_attribute(sf,l_string(std_cm,"flame.ButtonsControl"));
	if ( d )  {
	int id;
		id = atoi(n_string(std_cm,d));
		sts->buttons_control = id2marshaler(id);
		if ( sts->buttons_control == 0 ) {
			ret = "flame.ButtonsControl";
			goto err;
		}
		(*flags) |= VSF_FLAME_BUTTONS_CONTROL;
	}
	d = get_sf_attribute(sf,l_string(std_cm,"flame.IndicateObj"));
	if ( d )  {
		sts->indicate_obj = (VgbIndicate*)get_object_from_name(env,d);
		if ( sts->indicate_obj == 0 ) {
			ret = "flame.IndicateObj";
			goto err;
		}
		(*flags) |= VSF_FLAME_INDICATE_OBJ;
	}
	d = get_sf_attribute(sf,l_string(std_cm,"flame.IndicateTreeObj"));
	if ( d )  {
		sts->indicate_tree_obj = (VxlTreeView*)get_object_from_name(env,d);
		if ( sts->indicate_obj == 0 ) {
			ret = "flame.IndicateTreeObj";
			goto err;
		}
		(*flags) |= VSF_FLAME_INDICATE_TREE_OBJ;
	}
	d = get_sf_attribute(sf,l_string(std_cm,"flame.IndicateTreeEye.raw"));
	if ( d )  {
		sts->indicate_tree_eye = 
			get_image_from_lchar(&_ret,env,arg->h.file->name,d);
		if ( get_type(_ret) == XLT_ERROR ) {
			ret = "flame.IndicateTreeEye.raw";
			goto err;
		}
		(*flags) |= VSF_FLAME_INDICATE_TREE_EYE;
	}
	d = get_sf_attribute(sf,l_string(std_cm,"flame.IndicateTreeEye.elr0"));
	if ( d ) {
		sts->indicate_tree_eye_elr[0]
			= get_elr(n_string(std_cm,d),sts->indicate_tree_eye);
	}
	d = get_sf_attribute(sf,l_string(std_cm,"flame.IndicateTreeEye.elr1"));
	if ( d ) {
		sts->indicate_tree_eye_elr[1]
			= get_elr(n_string(std_cm,d),sts->indicate_tree_eye);
	}
	d = get_sf_attribute(sf,l_string(std_cm,"flame.IndicateTreeEye.elr"));
	if ( d ) {
		get_elr_ary(
			&sts->indicate_tree_eye_elr[0],
			n_string(std_cm,d),sts->indicate_tree_eye);
	}

	d = get_sf_attribute(sf,l_string(std_cm,"flame.IndicateTreeInfo.raw"));
	if ( d )  {
		sts->indicate_tree_info = 
			get_image_from_lchar(&_ret,env,arg->h.file->name,d);
		if ( get_type(_ret) == XLT_ERROR ) {
			ret = "flame.IndicateTreeInfo.raw";
			goto err;
		}
		(*flags) |= VSF_FLAME_INDICATE_TREE_INFO;
	}
	d = get_sf_attribute(sf,l_string(std_cm,"flame.IndicateTreeInfo.elr0"));
	if ( d ) {
		sts->indicate_tree_info_elr[0]
			= get_elr(n_string(std_cm,d),sts->indicate_tree_info);
	}
	d = get_sf_attribute(sf,l_string(std_cm,"flame.IndicateTreeInfo.elr1"));
	if ( d ) {
		sts->indicate_tree_info_elr[1]
			= get_elr(n_string(std_cm,d),sts->indicate_tree_info);
	}
	d = get_sf_attribute(sf,l_string(std_cm,"flame.IndicateTreeInfo.elr"));
	if ( d ) {
		get_elr_ary(
			&sts->indicate_tree_info_elr[0],
			n_string(std_cm,d),sts->indicate_tree_info);
	}

	d = get_sf_attribute(sf,l_string(std_cm,"flame.IndicateTreeBase.raw"));
	if ( d )  {
		sts->indicate_tree_base = 
			get_image_from_lchar(&_ret,env,arg->h.file->name,d);
		if ( get_type(_ret) == XLT_ERROR ) {
			ret = "flame.IndicateTreeBase.raw";
			goto err;
		}
		(*flags) |= VSF_FLAME_INDICATE_TREE_BASE;
	}
	d = get_sf_attribute(sf,l_string(std_cm,"flame.IndicateTreeBase.elr0"));
	if ( d ) {
		sts->indicate_tree_base_elr[0]
			= get_elr(n_string(std_cm,d),sts->indicate_tree_base);
	}
	d = get_sf_attribute(sf,l_string(std_cm,"flame.IndicateTreeBase.elr1"));
	if ( d ) {
		sts->indicate_tree_base_elr[1]
			= get_elr(n_string(std_cm,d),sts->indicate_tree_base);
	}
	d = get_sf_attribute(sf,l_string(std_cm,"flame.IndicateTreeBase.elr2"));
	if ( d ) {
		sts->indicate_tree_base_elr[2]
			= get_elr(n_string(std_cm,d),sts->indicate_tree_base);
	}
	d = get_sf_attribute(sf,l_string(std_cm,"flame.IndicateTreeBase.elr3"));
	if ( d ) {
		sts->indicate_tree_base_elr[3]
			= get_elr(n_string(std_cm,d),sts->indicate_tree_base);
	}
	d = get_sf_attribute(sf,l_string(std_cm,"flame.IndicateTreeBase.elr"));
	if ( d ) {
		get_elr_ary(
			&sts->indicate_tree_base_elr[0],
			n_string(std_cm,d),sts->indicate_tree_base);
	}

	d = get_sf_attribute(sf,l_string(std_cm,"flame.IndicateTreeWeb.raw"));
	if ( d )  {
		sts->indicate_tree_web = 
			get_image_from_lchar(&_ret,env,arg->h.file->name,d);
		if ( get_type(_ret) == XLT_ERROR ) {
			ret = "flame.IndicateTreeWeb.raw";
			goto err;
		}
		(*flags) |= VSF_FLAME_INDICATE_TREE_WEB;
	}
	d = get_sf_attribute(sf,l_string(std_cm,"flame.IndicateTreeWeb.elr0"));
	if ( d ) {
		sts->indicate_tree_web_elr[0]
			= get_elr(n_string(std_cm,d),sts->indicate_tree_web);
	}
	d = get_sf_attribute(sf,l_string(std_cm,"flame.IndicateTreeWeb.elr1"));
	if ( d ) {
		sts->indicate_tree_web_elr[1]
			= get_elr(n_string(std_cm,d),sts->indicate_tree_web);
	}
	d = get_sf_attribute(sf,l_string(std_cm,"flame.IndicateTreeWeb.elr2"));
	if ( d ) {
		sts->indicate_tree_web_elr[3]
			= get_elr(n_string(std_cm,d),sts->indicate_tree_web);
	}
	d = get_sf_attribute(sf,l_string(std_cm,"flame.IndicateTreeWeb.elr"));
	if ( d ) {
		get_elr_ary(
			&sts->indicate_tree_web_elr[0],
			n_string(std_cm,d),sts->indicate_tree_web);
	}

	d = get_sf_attribute(sf,l_string(std_cm,"flame.IndicateTreeAuth.raw"));
	if ( d )  {
		sts->indicate_tree_auth = 
			get_image_from_lchar(&_ret,env,arg->h.file->name,d);
		if ( get_type(_ret) == XLT_ERROR ) {
			ret = "flame.IndicateTreeAuth.raw";
			goto err;
		}
		(*flags) |= VSF_FLAME_INDICATE_TREE_AUTH;
	}
	d = get_sf_attribute(sf,l_string(std_cm,"flame.IndicateTreeAuth.elr0"));
	if ( d ) {
		sts->indicate_tree_auth_elr[0]
			= get_elr(n_string(std_cm,d),sts->indicate_tree_auth);
	}
	d = get_sf_attribute(sf,l_string(std_cm,"flame.IndicateTreeAuth.elr1"));
	if ( d ) {
		sts->indicate_tree_auth_elr[1]
			= get_elr(n_string(std_cm,d),sts->indicate_tree_auth);
	}
	d = get_sf_attribute(sf,l_string(std_cm,"flame.IndicateTreeAuth.elr2"));
	if ( d ) {
		sts->indicate_tree_auth_elr[3]
			= get_elr(n_string(std_cm,d),sts->indicate_tree_auth);
	}
	d = get_sf_attribute(sf,l_string(std_cm,"flame.IndicateTreeAuth.elr"));
	if ( d ) {
		get_elr_ary(
			&sts->indicate_tree_auth_elr[0],
			n_string(std_cm,d),sts->indicate_tree_auth);
	}

	d = get_sf_attribute(sf,l_string(std_cm,"flame.OAttributeHandler"));
	if ( d ) {
	VSysArgTranslator * vsat;
	VObject::FreeList * list;
		vsat = get_vsat(env,d);
		if ( vsat == 0 ) {
			ret = "OAttributeHandler";
			goto err;
		}
		vsat->translator = vsat_VgbFlameStandard_translator;
		sts->oattribute_handler = vobj_callback_glue;
		sts->oattribute_handler_arg = vsat;

		list = new VObject::FreeList;
		list->target = vsat;
		list->next = *free_list;
		list->free_func = free_VSysArgTranslator;
		insert_vsat_list(vsat);
		*free_list = list;
		(*flags) |= VSF_FLAME_OATTRIBUTE_HANDLER;
	}
	d = get_sf_attribute(sf,l_string(std_cm,"flame.SelectMode"));
	if ( d )  {
		if ( l_strcmp(d,l_string(std_cm,"GrabMode")) == 0 )
			sts->select_mode = VGB_MODE_GRAB;
		else if ( l_strcmp(d,l_string(std_cm,"RotateMode")) == 0 )
			sts->select_mode = VGB_MODE_ROTATE;
		else if ( l_strcmp(d,l_string(std_cm,"ClickMode")) == 0 )
			sts->select_mode = VGB_MODE_CLICK;
		else {
			ret = "flame.SelectMode";
			goto err;
		}
		(*flags) |= VSF_FLAME_SELECT_MODE;
	}

	d = get_sf_attribute(sf,l_string(std_cm,"flame.Focus"));
	if ( d ) {
		if ( l_strcmp(d,l_string(std_cm,"on")) == 0 )
			(*flags) |= VSF_FLAME_FOCUS;
	}

	return 0;
err:
	return (char*)ret;
}

char *
VgbFlame_get_flame_radar_from_sf(
	VgbFlameRadarStatus * sts,
	int * flags,
	XLISP_ENV * env,
	XL_SYM_FIELD * sf)
{
L_CHAR * d;
const char * ret;
XL_SEXP * _q;
	*flags = 0;
	ret = 0;
	memset(sts,0,sizeof(*sts));
	d = get_sf_attribute(sf,l_string(std_cm,"flame.RadarGetQuery"));
	if ( d ) {
		*flags |= VSF_FLAME_R_QUERY_GET;
	}
	d = get_sf_attribute(sf,l_string(std_cm,"flame.RadarInsertQuery"));
	if ( d ) {
		*flags |= VSF_FLAME_R_QUERY_INSERT;
		_q = eval(env,get_symbol(d));
		if ( get_type(_q) == XLT_ERROR ) {
			ret = "invalid query";
			goto err;
		}
		sts->query_insert = sexp_to_q_thread(_q);
	}
	d = get_sf_attribute(sf,l_string(std_cm,"flame.RadarSetQuery"));
	if ( d ) {
		*flags |= VSF_FLAME_R_QUERY_SET;
		_q = eval(env,get_symbol(d));
		if ( get_type(_q) == XLT_ERROR ) {
			ret = "invalid query";
			goto err;
		}
		sts->query_set = sexp_to_q_thread(_q);
	}
	d = get_sf_attribute(sf,l_string(std_cm,"flame.RadarReplaceQuery"));
	if ( d ) {
		*flags |= VSF_FLAME_R_QUERY_REPLACE;
		_q = eval(env,get_symbol(d));
		if ( get_type(_q) == XLT_ERROR ) {
			ret = "invalid query";
			goto err;
		}
		sts->query_replace = sexp_to_q_thread(_q);
	}
	d = get_sf_attribute(sf,l_string(std_cm,"flame.RadarDeleteQueryByNo"));
	if ( d ) {
		*flags |= VSF_FLAME_R_QUERY_DELETE_BYNO;
		if ( l_strcmp(d,l_string(std_cm,"all")) == 0 )
			sts->delete_byno = -1;
		else	sts->delete_byno = atoi(n_string(std_cm,d));
	}
	return 0;
err:
	return (char*)ret;
}

void
free_VgbFlameRadarStatus(VgbFlameRadarStatus * sts,int flags)
{
	if ( flags & VSF_FLAME_R_QUERY_GET )
		free_query_thread(sts->result_query);
	if ( flags & VSF_FLAME_R_QUERY_INSERT )
		free_query_thread(sts->query_insert);
}

XL_SEXP *
vobj_VgbFlameStandard(XLISP_ENV *env, XL_SEXP *arg, XLISP_ENV *a, XL_SYM_FIELD *sf)
{
VExError ex_er;
VObjectStatus _sts;
VgbFlameSwitchStatus sw_sts;
VgbFlameStatus f_sts;
VgbFlameRadarStatus r_sts;
VgbFlameCoordinateStatus c_sts;
int f_flags,sw_flags,r_flags,c_flags;
char * param_ret;
/*
int flags = get_sts_from_sf(&sts,0, sf, &free_list);
*/
XL_SEXP * p;
/*
L_CHAR * refer;
*/
VgbFlameStandard * obj;
VObject::FreeList * free_list = 0;
VObjectAppStatusAry app[2];

	memset(&f_sts,0,sizeof(f_sts));
	param_ret = VgbFlame_get_sw_from_sf(&sw_sts,&sw_flags,sf);
	if ( param_ret )
		goto param_err;
	param_ret = VgbFlame_get_flame_from_sf(&f_sts,&f_flags,env,sf,arg,&free_list);
	if ( param_ret )
		goto param_err;
	param_ret = VgbFlame_get_flame_radar_from_sf(&r_sts,&r_flags,env,sf);
	if ( param_ret ) {
		goto param_err;
	}
	param_ret = VgbFlame_get_flame_coordinate_from_sf(&c_sts,&c_flags,env,sf);
	if ( param_ret ) {
		goto param_err;
	}
	app[0].sts = (void*)&f_sts;
	app[0].flags = f_flags;
	app[1].sts = 0;
	app[1].flags = 0;
	p = get_refered_object<VgbFlameStandard>(&obj,env,arg,sf,
			VO_GFLM,"VgbFlameStandard",
			&_sts,0,app);
	if ( get_type(p) == XLT_ERROR )
		return p;

	obj->free_on_release_list(free_list);
	ex_er = obj->set_switch_status(&sw_sts,sw_flags);
	if ( ex_er.code != V_ER_NO_ERR )
		goto err2;
	if ( get_type(p) != XLT_INTEGER ) {
		ex_er = obj->set_flame_status(&f_sts,f_flags);
		if ( ex_er.code != V_ER_NO_ERR )
			goto err2;
	}
	ex_er = obj->set_radar_status(&r_sts,r_flags);
	free_VgbFlameRadarStatus(&r_sts,r_flags);
	if ( ex_er.code != V_ER_NO_ERR )
		goto err2;
	ex_er = obj->set_coordinate_status(&c_sts,c_flags);
	free_VgbFlameCoordinateStatus(&c_sts,c_flags);
	if ( ex_er.code != V_ER_NO_ERR )
		goto err2;
	if ( get_type(p) == XLT_INTEGER )
		return vobj_get_id_list(p->integer.data, 0, sf,0);
	else	return 0;
err2:
//	obj->destroy();
	if ( f_sts.indicate_tree_eye )
		v_image_unref(f_sts.indicate_tree_eye);
	if ( f_sts.indicate_tree_info )
		v_image_unref(f_sts.indicate_tree_info);
	if ( f_sts.indicate_tree_web )
		v_image_unref(f_sts.indicate_tree_web);

	return vobj_get_error(ex_er, arg,0);
param_err:
	if ( f_sts.indicate_tree_eye )
		v_image_unref(f_sts.indicate_tree_eye);
	if ( f_sts.indicate_tree_info )
		v_image_unref(f_sts.indicate_tree_info);
	if ( f_sts.indicate_tree_web )
		v_image_unref(f_sts.indicate_tree_web);

	return get_error(
		arg->h.file,
		arg->h.line,
		XLE_PROTO_INV_PARAM,
		l_string(std_cm,"VgbFlameStandard"),
		List(n_get_string("invalid param of attribute ::"),
			n_get_string(param_ret),
			-1));
}

extern "C" void
xx_free_wpm_1(WARP_POINT_MENU * m,char * f,int ln)
{
	if ( m->name )
		xx_d_f_ree(m->name,f,ln);
	if ( m->submenu )
		xx_free_wpm(m->submenu,f,ln);
	if ( m->wp )
		free_warp_point(m->wp);
	if ( m->path )
		xx_d_f_ree(m->path,f,ln);
	xx_d_f_ree(m,f,ln);
}

extern "C" void
xx_free_wpm(WARP_POINT_MENU* m,char * f,int ln)
{
WARP_POINT_MENU * m1;
	for ( ; m ; ) {
		m1 = m;
		m = m1->next;
		xx_free_wpm_1(m1,f,ln);
	}
}

extern "C" WARP_POINT_MENU*
copy_wpm(WARP_POINT_MENU* m)
{
WARP_POINT_MENU * ret, ** retp, * ret1;
	ret = 0;
	retp = &ret;
	for ( ; m ; m = m->next ) {
		ret1 = (WARP_POINT_MENU*)d_alloc(sizeof(*ret1));
		memset(ret1,0,sizeof(*ret1));
		*ret1 = *m;
		if ( m->name )
			ret1->name = ll_copy_str(m->name);
		if ( m->submenu )
			ret1->submenu = copy_wpm(m->submenu);
		if ( m->path )
			ret1->path = ll_copy_str(m->path);
		if ( m->wp )
			ret1->wp = rc_copy_warp_point(m->wp);
		ret1->next = 0;
		*retp = ret1;
		retp = &ret1->next;
	}
	return ret;
}

XL_SEXP *
wpm2sexp_1(WARP_POINT_MENU * m)
{
XL_SEXP * ret;
XL_SEXP * sym;
	ret = 0;
	if ( m->wp )
		ret = cons(wp2sexp(m->wp),ret);
	switch ( m->submenu_type ) {
	case WPM_ST_NONE:
		break;
	case WPM_ST_NORMAL:
		ret = cons(cons(n_get_symbol("submenu"),wpm2sexp(m->submenu,0)),ret);
		break;
	case WPM_ST_FILE:
		sym = n_get_symbol("submenu");
		set_attribute(sym,l_string(std_cm,"stype"),l_string(std_cm,"file"));
		set_attribute(sym,l_string(std_cm,"path"),m->path);
		ret = cons(cons(sym,wpm2sexp(m->submenu,0)),ret);
		break;
	case WPM_ST_NET:
		sym = n_get_symbol("submenu");
		set_attribute(sym,l_string(std_cm,"stype"),l_string(std_cm,"net"));
		set_attribute(sym,l_string(std_cm,"path"),m->path);
		ret = cons(cons(sym,wpm2sexp(m->submenu,0)),ret);
		break;
	default:
		return 0;
	}
	ret = cons(List(n_get_symbol("title"),get_string(m->name),-1),ret);
	return cons(n_get_symbol("menu-item"),ret);
}

XL_SEXP *
wpm2sexp(WARP_POINT_MENU * m,int tag_flag)
{
XL_SEXP * ret,*s;
	if ( tag_flag )
		ret = List(n_get_symbol("warp-point-menu"),-1);
	else	ret = 0;
	for ( ; m ; m = m->next ) {
		s = wpm2sexp_1(m);
		if ( s )
			ret = cons(s,ret);
	}
	return reverse(ret);
}

void
print_wpm(char * msg,WARP_POINT_MENU * w)
{
XL_SEXP * s;
	s = wpm2sexp(w,1);
	ss_printf("%s ",msg);
	print_sexp(s_stdout,s,PF_INDENT);
	ss_printf("\n");
}

WARP_POINT_MENU *
sexp2wpm_1(XL_SEXP * s)
{
XL_SEXP * target, * sym;
L_CHAR * stype;
WARP_POINT_MENU * m;
	m = (WARP_POINT_MENU*)d_alloc(sizeof(*m));
	memset(m,0,sizeof(*m));
	target = get_el_by_symbol(s,l_string(std_cm,"warp-point"),0);
	if ( target )
		m->wp = sexp2wp(target);
	target = get_el_by_symbol(s,l_string(std_cm,"submenu"),0);
	if ( target ) {
		sym = car(target);
		stype = get_sf_attribute(sym->symbol.field,l_string(std_cm,"stype"));
		m->path = ll_copy_str(get_sf_attribute(sym->symbol.field,l_string(std_cm,"path")));
		if ( stype == 0 || l_strcmp(stype,l_string(std_cm,"normal")) == 0 ) {
			m->submenu_type = WPM_ST_NORMAL;
			m->submenu = sexp2wpm(target);
		}
		else if ( l_strcmp(stype,l_string(std_cm,"file")) == 0 ) {
			m->submenu_type = WPM_ST_FILE;
			m->submenu = sexp2wpm(target);
		}
		else if ( l_strcmp(stype,l_string(std_cm,"net")) == 0 ) {
			m->submenu_type = WPM_ST_NET;
			m->submenu = sexp2wpm(target);
		}
		else {
			m->submenu_type = WPM_ST_NONE;
		}
	}
	target = get_el_by_symbol(s,l_string(std_cm,"title"),0);
	if ( target == 0 )
		goto err;
	target = get_el(target,1);
	if ( get_type(target) != XLT_STRING )
		goto err;
	m->name = ll_copy_str(target->string.data);
	if ( m->wp == 0 && m->submenu_type == WPM_ST_NONE )
		goto err;
	return m;
err:
	if ( m->wp )
		free_warp_point(m->wp);
	if ( m->submenu )
		free_wpm(m->submenu);
	if ( m->name )
		d_f_ree(m->name);
	return 0;
}

WARP_POINT_MENU *
sexp2wpm(XL_SEXP * s)
{
WARP_POINT_MENU * ret, ** mp,*m;
	ret = 0;
	mp = &ret;
	for ( ; get_type(s) == XLT_PAIR ; s = cdr(s) ) {
		if ( get_type(car(s)) != XLT_PAIR )
			continue;
		m = sexp2wpm_1(car(s));
		if ( m == 0 )
			continue;
		m->id = 0;
		*mp = m;
		mp = &m->next;
	}
	return ret;
}

XL_SEXP *
vobj_load_warp_point_menu(XLISP_ENV * env,XL_SEXP * s,XLISP_ENV * a,XL_SYM_FIELD * sf)
{
XL_SEXP * f;
L_CHAR * f_path;
char file[100];
L_CHAR * type;
L_CHAR * preference_filename;
int _type;
	type = get_sf_attribute(sf,l_string(std_cm,"type"));
	if ( type == 0 )
		_type = 1;
	else	_type = atoi(n_string(std_cm,type));
	preference_filename = get_sf_attribute(sf,l_string(std_cm,"preference"));
	sprintf(file,"%s%s",get_preference_path(),n_string(std_cm,preference_filename));
	if ( VgbFlame_load_wpm(l_string(&utf8_cm,file),_type,0) == 0 )
		return 0;
	if ( list_length(s) == 1 )
		goto cannot_open;
	f = get_el(s,1);
	if ( get_type(f) != XLT_STRING )
		goto type_missmatch;
	f_path = get_absolute_path(f->string.data);
	if ( VgbFlame_load_wpm(f_path,_type,1) < 0 )
		goto cannot_open;
	return 0;
type_missmatch:
	return get_error(
		s->h.file,
		s->h.line,
		XLE_SEMANTICS_TYPE_MISSMATCH,
		l_string(std_cm,"VgbFlameStandardLoadWPM"),
		n_get_string("file name type missmatch"));
cannot_open:
	return get_error(
		s->h.file,
		s->h.line,
		XLE_PROTO_OPEN_FILE,
		l_string(std_cm,"VgbFlameStandardLoadWPM"),
		n_get_string("preference and the given file can not be opened"));
}

XL_SEXP *
gb_trackback(XLISP_ENV * env,XL_SEXP * s,XLISP_ENV * a,XL_SYM_FIELD * sf)
{
XL_SEXP * inp,*sym;
XL_SEXP * ret;
RADAR_CACHE * rc_ptr;
WARP_POINT * wp;
const char * err;

ss_printf("TRACKBACK\n");
	inp = get_el(s,1);
	switch ( get_type(inp) ) {
	case XLT_STRING:
		rc_ptr = get_sf_rc_ptr(sf);
		if ( rc_trackback(&ret,rc_ptr,inp->string.data) < 0 ) {
			if ( get_type(ret) == XLT_ERROR )
				return ret;
			return get_error(
				s->h.file,
				s->h.line,
				XLE_PROTO_OPEN_FILE,
				l_string(std_cm,"trackback"),
				List(n_get_string("cannot open a window for trackback"),
					ret,-1));
		}
		break;
	case XLT_PAIR:
		sym = car(inp);
		if ( get_type(sym) != XLT_SYMBOL )
			goto type_missmatch;
		if ( l_strcmp(sym->symbol.data,l_string(std_cm,"warp-point")) == 0 ) {
			wp = sexp2wp(inp);
			if ( wp == 0 ) {
				err = "trackback warp-point type missmatch";
				goto type_missmatch;
			}
			rc_ptr = get_sf_rc_ptr(sf);
			if ( rc_ptr == 0 ) {
				return get_error(
					s->h.file,
					s->h.line,
					XLE_PROTO_OPEN_FILE,
					l_string(std_cm,"trackback"),
					List(n_get_string("cannot open a window for trackback"),
						-1));
			}
			rc_warp(rc_ptr,WT_WARP_POINT,0,0,0,0,0,wp,1);
			free_warp_point(wp);
		}
		else {
			err = "invalid symbol in trackback";
			goto type_missmatch;
		}
		break;
	default:
		err = "trackback URL type missmatch";
		goto type_missmatch;
	}
	return 0;
type_missmatch:
	return get_error(
		s->h.file,
		s->h.line,
		XLE_SEMANTICS_TYPE_MISSMATCH,
		l_string(std_cm,"trackback"),
		n_get_string(err));
}


} // extern "C"



